@@ -75,7 +75,7 @@ class Box:
 
     @low.setter
     def low(self, value):
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
+        v = numpy.array(value, ndmin=1, copy=True, dtype=numpy.float64)
         if v.shape != (3,):
             raise TypeError('Low must be a 3-tuple')
         self._low = v
@@ -87,7 +87,7 @@ class Box:
 
     @high.setter
     def high(self, value):
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
+        v = numpy.array(value, ndmin=1, copy=True, dtype=numpy.float64)
         if v.shape != (3,):
             raise TypeError('High must be a 3-tuple')
         self._high = v
@@ -101,7 +101,7 @@ class Box:
     def tilt(self, value):
         v = value
         if v is not None:
-            v = numpy.array(v, ndmin=1, copy=False, dtype=numpy.float64)
+            v = numpy.array(v, ndmin=1, copy=True, dtype=numpy.float64)
             if v.shape != (3,):
                 raise TypeError('Tilt must be a 3-tuple')
         self._tilt = v
@@ -130,6 +130,7 @@ class Snapshot:
         self._box = Box.cast(box)
         self.step = step
 
+        self._id = None
         self._position = None
         self._velocity = None
         self._image = None
@@ -148,6 +149,36 @@ class Snapshot:
         """:class:`Box`: Simulation box."""
         return self._box
 
+    @property
+    def id(self):
+        """:class:`numpy.ndarray`: Particle IDs."""
+        if not self.has_id():
+            self._id = numpy.arange(1, self.N+1)
+        return self._id
+
+    @id.setter
+    def id(self, value):
+        if value is not None:
+            v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.int32)
+            if v.shape != (self.N,):
+                raise TypeError('Ids must be a size N array')
+            if not self.has_id():
+                self._id = numpy.arange(1, self.N+1)
+            numpy.copyto(self._id, v)
+        else:
+            self._id = None
+
+    def has_id(self):
+        """Check if configuration has particle IDs.
+
+        Returns
+        -------
+        bool
+            True if particle IDs have been initialized.
+
+        """
+        return self._id is not None
+
     @property
     def position(self):
         """:class:`numpy.ndarray`: Positions."""
@@ -157,12 +188,15 @@ class Snapshot:
 
     @position.setter
     def position(self, value):
-        if not self.has_position():
-            self._position = numpy.zeros((self.N, 3), dtype=numpy.float64)
-        v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.float64)
-        if v.shape != (self.N, 3):
-            raise TypeError('Positions must be an Nx3 array')
-        numpy.copyto(self._position, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.float64)
+            if v.shape != (self.N, 3):
+                raise TypeError('Positions must be an Nx3 array')
+            if not self.has_position():
+                self._position = numpy.zeros((self.N, 3), dtype=numpy.float64)
+            numpy.copyto(self._position, v)
+        else:
+            self._position = None
 
     def has_position(self):
         """Check if configuration has positions.
@@ -184,12 +218,15 @@ class Snapshot:
 
     @image.setter
     def image(self, value):
-        if not self.has_image():
-            self._image = numpy.zeros((self.N, 3), dtype=numpy.int32)
-        v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.int32)
-        if v.shape != (self.N, 3):
-            raise TypeError('Images must be an Nx3 array')
-        numpy.copyto(self._image, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.int32)
+            if v.shape != (self.N, 3):
+                raise TypeError('Images must be an Nx3 array')
+            if not self.has_image():
+                self._image = numpy.zeros((self.N, 3), dtype=numpy.int32)
+            numpy.copyto(self._image, v)
+        else:
+            self._image = None
 
     def has_image(self):
         """Check if configuration has images.
@@ -211,12 +248,15 @@ class Snapshot:
 
     @velocity.setter
     def velocity(self, value):
-        if not self.has_velocity():
-            self._velocity = numpy.zeros((self.N, 3), dtype=numpy.float64)
-        v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.float64)
-        if v.shape != (self.N, 3):
-            raise TypeError('Velocities must be an Nx3 array')
-        numpy.copyto(self._velocity, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=2, copy=False, dtype=numpy.float64)
+            if v.shape != (self.N, 3):
+                raise TypeError('Velocities must be an Nx3 array')
+            if not self.has_velocity():
+                self._velocity = numpy.zeros((self.N, 3), dtype=numpy.float64)
+            numpy.copyto(self._velocity, v)
+        else:
+            self._velocity = None
 
     def has_velocity(self):
         """Check if configuration has velocities.
@@ -238,12 +278,15 @@ class Snapshot:
 
     @molecule.setter
     def molecule(self, value):
-        if not self.has_molecule():
-            self._molecule = numpy.zeros(self.N, dtype=numpy.int32)
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.int32)
-        if v.shape != (self.N,):
-            raise TypeError('Molecules must be a size N array')
-        numpy.copyto(self._molecule, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.int32)
+            if v.shape != (self.N,):
+                raise TypeError('Molecules must be a size N array')
+            if not self.has_molecule():
+                self._molecule = numpy.zeros(self.N, dtype=numpy.int32)
+            numpy.copyto(self._molecule, v)
+        else:
+            self._molecule = None
 
     def has_molecule(self):
         """Check if configuration has molecule tags.
@@ -265,12 +308,15 @@ class Snapshot:
 
     @typeid.setter
     def typeid(self, value):
-        if not self.has_typeid():
-            self._typeid = numpy.ones(self.N, dtype=numpy.int32)
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.int32)
-        if v.shape != (self.N,):
-            raise TypeError('Type must be a size N array')
-        numpy.copyto(self._typeid, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.int32)
+            if v.shape != (self.N,):
+                raise TypeError('Type must be a size N array')
+            if not self.has_typeid():
+                self._typeid = numpy.ones(self.N, dtype=numpy.int32)
+            numpy.copyto(self._typeid, v)
+        else:
+            self._typeid = None
 
     def has_typeid(self):
         """Check if configuration has types.
@@ -292,12 +338,15 @@ class Snapshot:
 
     @charge.setter
     def charge(self, value):
-        if not self.has_charge():
-            self._charge = numpy.zeros(self.N, dtype=numpy.float64)
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
-        if v.shape != (self.N,):
-            raise TypeError('Charge must be a size N array')
-        numpy.copyto(self._charge, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
+            if v.shape != (self.N,):
+                raise TypeError('Charge must be a size N array')
+            if not self.has_charge():
+                self._charge = numpy.zeros(self.N, dtype=numpy.float64)
+            numpy.copyto(self._charge, v)
+        else:
+            self._charge = None
 
     def has_charge(self):
         """Check if configuration has charges.
@@ -319,12 +368,15 @@ class Snapshot:
 
     @mass.setter
     def mass(self, value):
-        if not self.has_mass():
-            self._mass = numpy.ones(self.N, dtype=numpy.float64)
-        v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
-        if v.shape != (self.N,):
-            raise TypeError('Mass must be a size N array')
-        numpy.copyto(self._mass, v)
+        if value is not None:
+            v = numpy.array(value, ndmin=1, copy=False, dtype=numpy.float64)
+            if v.shape != (self.N,):
+                raise TypeError('Mass must be a size N array')
+            if not self.has_mass():
+                self._mass = numpy.ones(self.N, dtype=numpy.float64)
+            numpy.copyto(self._mass, v)
+        else:
+            self._mass = None
 
     def has_mass(self):
         """Check if configuration has masses.
@@ -337,6 +389,41 @@ class Snapshot:
         """
         return self._mass is not None
 
+    def reorder(self, order, check_order=True):
+        """Reorder the particles in place.
+
+        Parameters
+        ----------
+        order : list
+            New order of indexes.
+        check_order : bool
+            If true, validate the new ``order`` before applying it.
+
+        """
+        # sanity check the sorting order before applying it
+        if check_order and self.N > 1:
+            sorted_order = numpy.sort(order)
+            if (sorted_order[0] != 0 or sorted_order[-1] != self.N-1
+                or not numpy.all(sorted_order[1:] == sorted_order[:-1]+1)):
+                raise ValueError('New order must be an array from 0 to N-1')
+
+        if self.has_id():
+            self._id = self._id[order]
+        if self.has_position():
+            self._position = self._position[order]
+        if self.has_velocity():
+            self._velocity = self._velocity[order]
+        if self.has_image():
+            self._image = self._image[order]
+        if self.has_molecule():
+            self._molecule = self._molecule[order]
+        if self.has_typeid():
+            self._typeid = self._typeid[order]
+        if self.has_charge():
+            self._charge = self._charge[order]
+        if self.has_mass():
+            self._mass = self._mass[order]
+
 class DataFile:
     """LAMMPS data file.
 
@@ -464,7 +551,7 @@ class DataFile:
             f.write("\nAtoms # {}\n\n".format(style))
             for i in range(snapshot.N):
                 style_args = dict(
-                    atomid=i+1,
+                    atomid=snapshot.id[i] if snapshot.has_id() else i+1,
                     typeid=snapshot.typeid[i],
                     x=snapshot.position[i][0],
                     y=snapshot.position[i][1],
@@ -485,7 +572,7 @@ class DataFile:
                 for i in range(snapshot.N):
                     vel_fmt="{atomid:8d}{vx:16.8f}{vy:16.8f}{vz:16.8f}\n"
                     f.write(vel_fmt.format(
-                        atomid=i+1,
+                        atomid=snapshot.id[i] if snapshot.has_id() else i+1,
                         vx=snapshot.velocity[i][0],
                         vy=snapshot.velocity[i][1],
                         vz=snapshot.velocity[i][2]))
@@ -494,7 +581,7 @@ class DataFile:
             if masses is not None:
                 f.write("\nMasses\n\n")
                 for i,mi in enumerate(masses):
-                    f.write("{typeid:4d}{m:12}\n".format(typeid=i+1,m=mi))
+                    f.write("{typeid:4d}{m:12}\n".format(typeid=i+1, m=mi))
 
         return DataFile(filename)
 
@@ -578,6 +665,7 @@ class DataFile:
                 raise IOError('Box bounds not read')
             box = Box(box_bounds[:3], box_bounds[3:], box_tilt)
             snap = Snapshot(N,box)
+            id_map = {}
 
             # now that snapshot is made, file it in with body sections
             masses = None
@@ -619,7 +707,17 @@ class DataFile:
                         # check that row is correctly sized and process
                         if not (len(row) == style_cols+4 or len(row) == style_cols+7):
                             raise IOError('Expected number of columns not read for atom style')
-                        idx = int(row[0])-1
+
+                        # only save the atom id if it is not in standard order
+                        id_ = int(row[0])
+                        if id_ != i+1:
+                            if id_ not in id_map:
+                                id_map[id_] = i
+                            idx = id_map[id_]
+                            snap.id[idx] = id_
+                        else:
+                            idx = i
+
                         if style == 'full':
                             snap.molecule[idx] = int(row[1])
                             snap.typeid[idx] = int(row[2])
@@ -645,7 +743,15 @@ class DataFile:
                         row = _readline(f,True).split()
                         if len(row) < 4:
                             raise IOError('Expected number of columns not read for velocity')
-                        idx = int(row[0])-1
+                        # parse atom id: need to repeat mapping in case Velocity comes before Atoms
+                        id_ = int(row[0])
+                        if id_ != i+1:
+                            if id_ not in id_map:
+                                id_map[id_] = i
+                            idx = id_map[id_]
+                            snap.id[idx] = id_
+                        else:
+                            idx = i
                         snap.velocity[idx] = [float(x) for x in row[1:4]]
                 elif 'Masses' in line:
                     masses = {}
@@ -674,9 +780,9 @@ class DumpFile:
     The dump file is a flexible file format, so a ``schema`` needs to be given
     to parse the atom data. The ``schema`` is given as a dictionary of column
     indexes. Valid keys for the schema match the names and shapes in the `Snapshot`.
-    The keys requiring only 1 column index are: `id`, `typeid`, `molecule`, `charge`,
-    and `mass`. The keys requiring 3 column indexes are `position`, `velocity`,
-    and `image`.
+    The keys requiring only 1 column index are: ``id``, ``typeid``, ``molecule``,
+    ``charge``, and ``mass``. The keys requiring 3 column indexes are ``position``,
+    ``velocity``, and ``image``.
 
     Parameters
     ----------
@@ -684,13 +790,21 @@ class DumpFile:
         Path to dump file.
     schema : dict
         Schema for the contents of the file.
+    sort_ids : bool
+        If true, sort the particles by ID in each snapshot.
+    copy_from : :class:`Snapshot`
+        If specified, copy fields that are missing in the dump file but are set in
+        a reference :class:`Snapshot`. The fields that can be copied are ``typeid``,
+        ``molecule``, ``charge``, and ``mass``.
 
     """
 
-    def __init__(self, filename, schema):
+    def __init__(self, filename, schema, sort_ids=True, copy_from=None):
         self.filename = filename
         self.schema = schema
         self._frames = None
+        self.sort_ids = sort_ids
+        self.copy_from = copy_from
 
     @classmethod
     def create(cls, filename, schema, snapshots):
@@ -764,7 +878,7 @@ class DumpFile:
                     line = ''
                     for col, (key, key_idx) in dump_row:
                         if key == 'id':
-                            val = i + 1
+                            val = snap.id[i] if snap.has_id() else i+1
                         else:
                             val = getattr(snap, key)[i]
                             if key_idx is not None:
@@ -792,6 +906,16 @@ class DumpFile:
 
         return DumpFile(filename, schema)
 
+    @property
+    def copy_from(self):
+        return self._copy_from
+
+    @copy_from.setter
+    def copy_from(self, value):
+        if value is not None and not isinstance(value, Snapshot):
+            raise TypeError('Dump file can only copy from Snapshot')
+        self._copy_from = value
+
     @property
     def filename(self):
         """str: Path to the file."""
@@ -909,27 +1033,54 @@ class DumpFile:
                         atom = atom.split()
 
                         if 'id' in self.schema:
-                            tag = int(atom[self.schema['id']]) - 1
-                        else:
-                            tag = i
-
+                            id_ = int(atom[self.schema['id']])
+                            if id_ != i+1:
+                                snap.id[i] = id_
                         if 'position' in self.schema:
-                            snap.position[tag] = [float(atom[j]) for j in self.schema['position']]
+                            snap.position[i] = [float(atom[j]) for j in self.schema['position']]
                         if 'velocity' in self.schema:
-                            snap.velocity[tag] = [float(atom[j]) for j in self.schema['velocity']]
+                            snap.velocity[i] = [float(atom[j]) for j in self.schema['velocity']]
                         if 'image' in self.schema:
-                            snap.image[tag] = [int(atom[j]) for j in self.schema['image']]
+                            snap.image[i] = [int(atom[j]) for j in self.schema['image']]
                         if 'molecule' in self.schema:
-                            snap.molecule[tag] = int(atom[self.schema['molecule']])
+                            snap.molecule[i] = int(atom[self.schema['molecule']])
                         if 'typeid' in self.schema:
-                            snap.typeid[tag] = int(atom[self.schema['typeid']])
+                            snap.typeid[i] = int(atom[self.schema['typeid']])
                         if 'charge' in self.schema:
-                            snap.charge[tag] = float(atom[self.schema['charge']])
+                            snap.charge[i] = float(atom[self.schema['charge']])
                         if 'mass' in self.schema:
-                            snap.mass[tag] = float(atom[self.schema['mass']])
+                            snap.mass[i] = float(atom[self.schema['mass']])
 
                 # final processing stage for the frame
                 if state == 4:
+                    # optionally sort the particles by ID
+                    if self.sort_ids and snap.has_id():
+                        snap.reorder(numpy.argsort(snap.id), check_order=False)
+
+                    # optionally copy reference data by ID / index
+                    if self._copy_from is not None:
+                        if snap.N != self._copy_from.N:
+                            raise ValueError('Cannot copy from a Snapshot with a different size')
+
+                        if self._copy_from.has_id():
+                            copy_id_map = {id_: i for i, id_ in enumerate(self._copy_from.id)}
+                        else:
+                            copy_id_map = {i+1: i for i in range(self._copy_from.N)}
+
+                        if snap.has_id():
+                            copy_id = [copy_id_map[id_] for id_ in snap.id]
+                        else:
+                            copy_id = [copy_id_map[id_] for id_ in range(1, snap.N+1)]
+
+                        if not snap.has_typeid() and self._copy_from.has_typeid():
+                            snap.typeid = self._copy_from.typeid[copy_id]
+                        if not snap.has_molecule() and self._copy_from.has_molecule():
+                            snap.molecule = self._copy_from.molecule[copy_id]
+                        if not snap.has_charge() and self._copy_from.has_charge():
+                            snap.charge = self._copy_from.charge[copy_id]
+                        if not snap.has_mass() and self._copy_from.has_mass():
+                            snap.mass = self._copy_from.mass[copy_id]
+
                     yield snap
                     del snap,N,box,step
                     state = 0
